#include "mbed.h"
#include "GPS_interrupt.h"

Serial pc(USBTX,USBRX);

Serial gps_bus(p9,p10);
GPS_interrupt gps(&gps_bus,9600);

double tx=130.216004,ty=33.594562;             //ゴール座標
double elat = 111319.528;           //緯度一度当たりの距離(m)
double elon = 91904.295;        //種子島における経度一度当たりの距離(m)

double pi=3.141592;

InterruptIn fpin(p25);
void up(); 

PwmOut servo(p22);

int a=1,b=1,c=1;

Serial sd(p28,p27);

Serial XB(p13,p14,115200);

DigitalOut mos(p21);
//------------------------------------------------------------------------------
//  超音波センサ
//------------------------------------------------------------------------------
Timer tmr;
DigitalOut trig(p23);
DigitalIn echo(p24);
int Otime,Atime;            //停止時間,開始時間
float Delta,ALT;            //開始時間と停止時間の差,地面からの高さ

    //------------------------------------------------------------------------------
    //  超音波センサを用いて機体の地面からの高さを測る関数
    //------------------------------------------------------------------------------
    void measureH(){
        trig=1;
        wait(0.00002);
        trig=0;
        while(echo!=1){}
        Atime=tmr.read_us();
        while(echo==1){}
        Otime=tmr.read_us();
        Delta=Otime-Atime;
        ALT=((Delta)/2)*340*1000/1000000;
        sd.printf("ALT = %f [mm],",ALT);
        wait(0.01);
    }

//------------------------------------------------------------------------------
//  移動
//------------------------------------------------------------------------------
DigitalOut dgo_R1(p18);
DigitalOut dgo_R2(p17);     //p17と18は入れ替わる可能性あり
DigitalOut dgo_L1(p19);
DigitalOut dgo_L2(p20);     //p19と20は入れ替わる可能性あり

    //前進
    void forward(){
        dgo_R1=0;
        dgo_R2=1;
        wait(0.3);
        dgo_R1=0;
        dgo_R2=1;
        dgo_L1=0;
        dgo_L2=1;
        XB.printf("Forward\r\n");
        sd.printf("Forward,");
    }

    //後退
    void back(){
        dgo_R1=1;
        dgo_R2=0;
        dgo_L1=1;
        dgo_L2=0;
        XB.printf("Buck\r\n");
        sd.printf("Buck,");
    }

    //停止
    void stop(){
        dgo_R1=1;
        dgo_R2=1;
        dgo_L1=1;
        dgo_L2=1;
        wait(0.50);
        dgo_R1=0;
        dgo_R2=0;
        dgo_L1=0;
        dgo_L2=0;
        XB.printf("Stop\r\n");
        sd.printf("Stop,");
    }

    //右折
    void right(){
        dgo_L1=0;
        dgo_L2=1;
        XB.printf("Turn right\r\n");
        sd.printf("Turn right,");
    }

    //左折
    void left(){
        dgo_R1=0;
        dgo_R2=1;
        XB.printf("Turn left\r\n");
        sd.printf("Turn left,");
    }

void up(){
    a=2;
}

//------------------------------------------------------------------------------
//  転倒回復
//------------------------------------------------------------------------------
void recover(){
    for(float p=0;p<0.001;p+=0.00001){
        servo.pulsewidth(0.001+p);
        wait(0.01);
    }
    for(float p=0.001;p>0;p-=0.00001){
        servo.pulsewidth(0.001+p);
        wait(0.01);
    }
    XB.printf("Recovering.\r\n");
    sd.printf("Recovering,");
}

//------------------------------------------------------------------------------
//  ここからメイン文
//------------------------------------------------------------------------------
int main(){

//------------------------------------------------------------------------------
//  フライトピン外れるまでの処理
//------------------------------------------------------------------------------
fpin.mode(PullUp);
    while(a==1){
    XB.printf("Stand by...\r\n");
    //pc.printf("Stand by...\r\n");
    sd.printf("Stand by...\r\n");
    float utc[6]={0};
    gps.getUTC(utc);
    XB.printf("%d/%d/%d %d:%d:%02.2f\r\n",(int)utc[0],(int)utc[1], (int)utc[2], (int)utc[3], (int)utc[4] ,utc[5]);
    //pc.printf("%d/%d/%d %d:%d:%02.2f\r\n",(int)utc[0],(int)utc[1], (int)utc[2], (int)utc[3], (int)utc[4] ,utc[5]);
    sd.printf("%d/%d/%d %d:%d:%02.2f\r\n",(int)utc[0],(int)utc[1], (int)utc[2], (int)utc[3], (int)utc[4] ,utc[5]);
    wait(3.0);
    fpin.rise(&up);
    }

//------------------------------------------------------------------------------
//  ここからフライトピン外れた後の処理
//------------------------------------------------------------------------------
XB.printf("\n");
sd.printf("\n");
XB.printf("Operation start.\r\n");
sd.printf("Operation start.\r\n");

tmr.start();
    while(b==1){
        measureH();
        XB.printf("ALT = %f [mm]\r\n",ALT);
        if(ALT>=10 && ALT<=2000){
            b=2;
        }
        else{
            if(Otime>60000000){
                b=2;
            }
        }
    }

sd.printf("\r\n");
XB.printf("Now separate Para.\r\n");
sd.printf("Now separate Para.\r\n");

mos=1;
wait(3.0);
mos=0;

wait(10.0);

//------------------------------------------------------------------------------
//  着地後のXBとGPSのチェック
//------------------------------------------------------------------------------
float utc[6]={0};
gps.getUTC(utc);
XB.printf("The Tank Has Landed at %d/%d/%d %d:%d:%02.2f \r\n",(int)utc[0],(int)utc[1], (int)utc[2], (int)utc[3], (int)utc[4] ,utc[5]);
sd.printf("The Tank Has Landed at %d/%d/%d %d:%d:%02.2f \r\n",(int)utc[0],(int)utc[1], (int)utc[2], (int)utc[3], (int)utc[4] ,utc[5]);        

forward();
wait(3.0);
stop();


    
    while(1){
        sd.printf("\r\n");
    //----------------------------------------------------------------------
    //  移動前の座標測定
    //----------------------------------------------------------------------
        double xy1[2]={0};
        float utc[6]={0};        
        gps.getPosition(xy1);
        gps.getUTC(utc);
        sd.printf("%d/%d/%d %d:%d:%02.2f,",(int)utc[0],(int)utc[1], (int)utc[2], (int)utc[3], (int)utc[4] ,utc[5]);
        XB.printf("Longitude:%3.7f,Latitude:%3.7f\r\n",xy1[0],xy1[1]);
        sd.printf("Longitude:%3.7f,Latitude:%3.7f,",xy1[0],xy1[1]);
        
        double X1=(xy1[0]-tx)*elon;         //目標までの東西方向の距離
        double Y1=(xy1[1]-ty)*elat;         //目標までの南北方向の距離
        double dis1=sqrt(X1*X1+Y1*Y1);      //移動前の座標から目標までの距離
        
        XB.printf("%3.2f[m] to target.\r\n",dis1);
        sd.printf("%3.2f[m] to target,",dis1);
        //------------------------------------------------------------------
        //  目標が近い場合
        //------------------------------------------------------------------
            if(dis1<5){
                forward();
                wait(5);
                stop();
                sd.printf("\r\n");
                XB.printf("Mission accomplished!!\r\n");
                sd.printf("Mission accomplished!!\r\n");
                break;
            }
            
        //----------------------------------------------------------------------
        //  目標まで距離がある場合
        //----------------------------------------------------------------------
            else{    
            //----------------------------------------------------------------------
            //  前進し、速度、角度を測定
            //----------------------------------------------------------------------
                forward();
                wait(5.0);
                double spd=gps.Knot()*1852/3600;    //機体の速度
                double deg=gps.Degree();            //機体の進行方向（北から右回り正）
                wait(5.0);
                stop();
    
            //----------------------------------------------------------------------
            //  移動後の座標測定
            //----------------------------------------------------------------------
                double xy2[2]={0};
                gps.getPosition(xy2);
                XB.printf("Longitude:%3.7f,Latitude:%3.7f\r\n",xy2[0],xy2[1]);
                sd.printf("Longitude:%3.7f,Latitude:%3.7f,",xy2[0],xy2[1]);
                sd.printf("%.3f[m/s], %3.2f[degree],",spd,deg);

                wait(0.1);
    
            //--------------------------------------------------------------------------
            //  距離の計算
            //--------------------------------------------------------------------------
                double X2=(xy2[0]-tx)*elon;      
                double Y2=(xy2[1]-ty)*elat;      
                double dis2=sqrt(X2*X2+Y2*Y2);      //移動後の座標から目標までの距離
        
                double X3=(xy1[0]-xy2[0])*elon;      
                double Y3=(xy1[1]-xy2[1])*elat;      
                double dis3=sqrt(X3*X3+Y3*Y3);      //移動距離
                measureH();
                
            //------------------------------------------------------------------
            //  移動後の座標と目標との角度
            //------------------------------------------------------------------
                double an=atan2(Y2,X2)*180/pi;             //atan(y,x) 

/*
                //--------------------------------------------------------------
                //  0.1m以上前進していない場合
                //--------------------------------------------------------------
                if(dis3<1){
                    while(1){
                        measureH();
                        //------------------------------------------------------
                        //  転倒していない場合
                        //------------------------------------------------------
                        if(0<ALT<100){
                            back();
                            wait(5);
                            stop();
                            c=2;
                        }
                        //------------------------------------------------------
                        //  転倒している場合
                        //------------------------------------------------------
                        else{
                            forward();
                            wait(2);
                            recover();
                            stop();
                        }
                    }
                }
*/
                if(100<ALT){
                    while(1){
                        XB.printf("Stumbled...\r\n");
                        sd.printf("Stumbled...,");
                        forward();
                        wait(1);
                        recover();
                        wait(1);
                        stop();
                        wait(3);
                        measureH();
                        if(0<ALT<100){
                            break;
                        }
                    }
                }
                
                //--------------------------------------------------------------
                //  1m以上進んでいる場合
                //--------------------------------------------------------------
                else{
                    //----------------------------------------------------------     
                    //  角度の計算(余弦定理）
                    //----------------------------------------------------------
                    double dir=180-acos((dis2*dis2+dis3*dis3-dis1*dis1)/(2*dis2*dis3))*180/pi ;       //回転角
                    XB.printf("Direction : %3.2f[deg]\r\n",dir);        
                    sd.printf("Direction : %3.2f[deg],",dir);
                    
                    //----------------------------------------------------------
                    //  機体を回転
                    //----------------------------------------------------------
                    double t=dir/60;         //角度[deg]÷回転速度[deg/s](仮に1)

                        if(xy2[0]>=tx&&xy2[1]>=ty){
                            if(0<=deg-(90-an)&&deg-(90-an)<180){       
                                right();
                                wait(t);
                                XB.printf("%3.2f[s]\r\n",t);
                                sd.printf("%3.2f[s],",t);
                                stop();
                            }
                            else{
                                left();
                                wait(t);
                                XB.printf("%3.2f[s]\r\n",t);
                                sd.printf("%3.2f[s],",t);
                                stop();
                            }    
                        }
                        
                        if(xy2[0]<tx&&xy2[1]>=ty){
                            if(0<=deg-(180-(an-90))&&deg-(180-(an-90))<180){
                                left();
                                wait(t);
                                XB.printf("%3.2f[s]\r\n",t);
                                sd.printf("%3.2f[s],",t);
                                stop();
                            }
                            else{
                                right();
                                wait(t);
                                XB.printf("%3.2f[s]\r\n",t);
                                sd.printf("%3.2f[s],",t);
                                stop();
                            }
                        }
                        
                        if(xy2[0]>=tx&&xy2[1]<ty){
                            if(0<=deg-(180-(90+an))&&deg-(180-(90+an))<180){
                                right();
                                wait(t);
                                XB.printf("%3.2f[s]\r\n",t);
                                sd.printf("%3.2f[s],",t);
                                stop();
                            }
                            else{
                                left();
                                wait(t);
                                XB.printf("%3.2f[s]\r\n",t);
                                sd.printf("%3.2f[s],",t);
                                stop();
                            }
                        }
                                
                                
                        if(xy2[0]<tx&&xy2[1]<ty){
                            if(0<=deg-(90-(180+an))&&deg-(90-(180+an))<180){
                                left();
                                wait(t);
                                XB.printf("%3.2f[s]\r\n",t);
                                sd.printf("%3.2f[s],",t);
                                stop();
                            }
                            else{
                                right();
                                wait(t);
                                XB.printf("%3.2f[s]\r\n",t);
                                sd.printf("%3.2f[s],",t);
                                stop();
                            }
                        }
                }
            }
    }
}
